import DeepProxy from "proxy-deep";
import { useQueryClient, createQuery, createMutation, createInfiniteQuery, createQueries, } from "@tanstack/svelte-query";
import { onDestroy } from "svelte";
// createUtils
const UtilsProcedureNames = {
    client: "client",
    fetch: "fetch",
    prefetch: "prefetch",
    fetchInfinite: "fetchInfinite",
    prefetchInfinite: "prefetchInfinite",
    ensureData: "ensureData",
    invalidate: "invalidate",
    refetch: "refetch",
    reset: "reset",
    cancel: "cancel",
    setData: "setData",
    getData: "getData",
    setInfiniteData: "setInfiniteData",
    getInfiniteData: "getInfiniteData",
};
// Procedures
const ProcedureNames = {
    query: "createQuery",
    serverQuery: "createServerQuery",
    infiniteQuery: "createInfiniteQuery",
    serverInfiniteQuery: "createServerInfiniteQuery",
    mutate: "createMutation",
    subscribe: "createSubscription",
    queryKey: "getQueryKey",
    context: "createContext",
    utils: "createUtils",
    queries: "createQueries",
    serverQueries: "createServerQueries",
};
// Implementation
function createQueriesProxy(client) {
    return new DeepProxy({}, {
        get() {
            return this.nest(() => { });
        },
        apply(_target, _thisArg, argList) {
            const target = [...this.path].reduce((client, value) => client[value], client);
            const [input, opts] = argList;
            return {
                ...opts,
                queryKey: getArrayQueryKey(this.path, input, "query"),
                queryFn: () => target.query(input),
            };
        },
    });
}
const utilsProcedures = {
    [UtilsProcedureNames.fetch]: ({ path, queryClient, target }) => {
        return (input, opts) => {
            return queryClient.fetchQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, "query"),
                queryFn: () => target.query(input),
            });
        };
    },
    [UtilsProcedureNames.prefetch]: ({ path, queryClient, target }) => {
        return (input, opts) => {
            return queryClient.prefetchQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, "query"),
                queryFn: () => target.query(input),
            });
        };
    },
    [UtilsProcedureNames.fetchInfinite]: ({ path, queryClient, target }) => {
        return (input, opts) => {
            return queryClient.fetchInfiniteQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, "infinite"),
                queryFn: ({ pageParam }) => target.query({ ...input, cursor: pageParam }),
            });
        };
    },
    [UtilsProcedureNames.prefetchInfinite]: ({ path, queryClient, target }) => {
        return (input, opts) => {
            return queryClient.prefetchInfiniteQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, "infinite"),
                queryFn: ({ pageParam }) => target.query({ ...input, cursor: pageParam }),
            });
        };
    },
    [UtilsProcedureNames.ensureData]: ({ path, queryClient, target }) => {
        return (input, opts) => {
            return queryClient.ensureQueryData({
                ...opts,
                queryKey: getArrayQueryKey(path, input, "query"),
                queryFn: () => target.query(input),
            });
        };
    },
    [UtilsProcedureNames.invalidate]: ({ path, queryClient }) => {
        return (input, filters, options) => {
            return queryClient.invalidateQueries({
                ...filters,
                queryKey: getArrayQueryKey(path, input, "any"),
            }, options);
        };
    },
    [UtilsProcedureNames.refetch]: ({ path, queryClient }) => {
        return (input, filters, options) => {
            return queryClient.refetchQueries({
                ...filters,
                queryKey: getArrayQueryKey(path, input, "any"),
            }, options);
        };
    },
    [UtilsProcedureNames.cancel]: ({ path, queryClient }) => {
        return (input, filters, options) => {
            return queryClient.cancelQueries({
                ...filters,
                queryKey: getArrayQueryKey(path, input, "any"),
            }, options);
        };
    },
    [UtilsProcedureNames.reset]: ({ queryClient, path }) => {
        return (input, filters, options) => {
            return queryClient.resetQueries({
                ...filters,
                queryKey: getArrayQueryKey(path, input, "any"),
            }, options);
        };
    },
    [UtilsProcedureNames.setData]: ({ queryClient, path }) => {
        return (input, updater, options) => {
            return queryClient.setQueryData(getArrayQueryKey(path, input, "query"), updater, options);
        };
    },
    [UtilsProcedureNames.setInfiniteData]: ({ queryClient, path }) => {
        return (input, updater, options) => {
            return queryClient.setQueryData(getArrayQueryKey(path, input, "infinite"), updater, options);
        };
    },
    [UtilsProcedureNames.getData]: ({ queryClient, path }) => {
        return (input) => {
            return queryClient.getQueryData(getArrayQueryKey(path, input, "query"));
        };
    },
    [UtilsProcedureNames.getInfiniteData]: ({ queryClient, path }) => {
        return (input) => {
            return queryClient.getQueryData(getArrayQueryKey(path, input, "infinite"));
        };
    },
};
function createUtilsProxy(client, queryClient) {
    return new DeepProxy({}, {
        get(_target, key, _receiver) {
            if (key === UtilsProcedureNames.client)
                return client;
            if (Object.hasOwn(utilsProcedures, key)) {
                const target = [...this.path].reduce((client, value) => client[value], client);
                return utilsProcedures[key]({ path: this.path, target, queryClient });
            }
            return this.nest(() => { });
        },
    });
}
const procedures = {
    [ProcedureNames.queryKey]: ({ path }) => {
        return (input, opts) => getArrayQueryKey(path, input, opts);
    },
    [ProcedureNames.query]: ({ path, target, abortOnUnmount }) => {
        const targetFn = target.query;
        return (input, opts) => {
            const shouldAbortOnUnmount = opts?.trpc?.abortOnUnmount ?? abortOnUnmount;
            return createQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, "query"),
                queryFn: ({ signal }) => targetFn(input, {
                    ...(shouldAbortOnUnmount && { signal }),
                }),
            });
        };
    },
    [ProcedureNames.serverQuery]: ({ path, target, queryClient, abortOnUnmount, }) => {
        const targetFn = target.query;
        return async (input, opts) => {
            const shouldAbortOnUnmount = opts?.trpc?.abortOnUnmount ?? abortOnUnmount;
            const query = {
                queryKey: getArrayQueryKey(path, input, "query"),
                queryFn: ({ signal }) => targetFn(input, {
                    ...(shouldAbortOnUnmount && { signal }),
                }),
            };
            const cache = queryClient
                .getQueryCache()
                .find({ queryKey: query.queryKey });
            const cacheNotFound = !cache?.state?.data;
            if (opts?.ssr !== false && cacheNotFound) {
                await queryClient.prefetchQuery(query);
            }
            const defaultOptions = queryClient.getDefaultOptions();
            return (...args) => {
                let newQuery = query;
                if (args.length > 0) {
                    const newInput = args[0];
                    let i = newInput;
                    if (typeof newInput === "function")
                        i = newInput(input);
                    newQuery = {
                        queryKey: getArrayQueryKey(path, i, "query"),
                        queryFn: ({ signal }) => targetFn(i, {
                            ...(shouldAbortOnUnmount && { signal }),
                        }),
                    };
                }
                return createQuery({
                    ...opts,
                    ...newQuery,
                    ...(cacheNotFound
                        ? {
                            refetchOnMount: opts?.refetchOnMount ??
                                defaultOptions.queries?.refetchOnMount ??
                                false,
                        }
                        : {}),
                });
            };
        };
    },
    [ProcedureNames.infiniteQuery]: ({ path, target, abortOnUnmount }) => {
        return (input, opts) => {
            const shouldAbortOnUnmount = opts?.trpc?.abortOnUnmount ?? abortOnUnmount;
            return createInfiniteQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, "infinite"),
                queryFn: ({ pageParam, signal }) => target.query({
                    ...input,
                    cursor: pageParam ?? opts?.initialCursor,
                    ...(shouldAbortOnUnmount && { signal }),
                }),
            });
        };
    },
    [ProcedureNames.serverInfiniteQuery]: ({ path, target, queryClient, abortOnUnmount, }) => {
        const targetFn = target.query;
        return async (input, opts) => {
            const shouldAbortOnUnmount = opts?.trpc?.abortOnUnmount ?? abortOnUnmount;
            const query = {
                queryKey: getArrayQueryKey(path, input, "infinite"),
                queryFn: ({ pageParam, signal }) => targetFn({
                    ...input,
                    cursor: pageParam ?? opts?.initialCursor,
                    ...(shouldAbortOnUnmount && { signal }),
                }),
            };
            const cache = queryClient
                .getQueryCache()
                .find({ queryKey: query.queryKey });
            const cacheNotFound = !cache?.state?.data;
            if (opts?.ssr !== false && cacheNotFound) {
                await queryClient.prefetchInfiniteQuery(query);
            }
            return (...args) => {
                let newQuery = query;
                if (args.length > 0) {
                    const newInput = args[0];
                    let i = newInput;
                    if (typeof newInput === "function")
                        i = newInput(input);
                    newQuery = {
                        queryKey: getArrayQueryKey(path, i, "infinite"),
                        queryFn: ({ pageParam, signal }) => targetFn({
                            ...i,
                            cursor: pageParam ?? opts?.initialCursor,
                            ...(shouldAbortOnUnmount && { signal }),
                        }),
                    };
                }
                const defaultOptions = queryClient.getDefaultOptions();
                return createInfiniteQuery({
                    ...opts,
                    ...newQuery,
                    ...(cacheNotFound
                        ? {
                            refetchOnMount: opts?.refetchOnMount ??
                                defaultOptions.queries?.refetchOnMount ??
                                false,
                        }
                        : {}),
                });
            };
        };
    },
    [ProcedureNames.mutate]: ({ path, target }) => {
        return (opts) => {
            return createMutation({
                ...opts,
                mutationKey: path,
                mutationFn: (data) => target.mutate(data),
            });
        };
    },
    [ProcedureNames.subscribe]: ({ target }) => {
        return (input, opts) => {
            const enabled = opts?.enabled ?? true;
            if (!enabled)
                return;
            let isStopped = false;
            const subscription = target.subscribe(input, {
                onStarted: () => {
                    if (!isStopped)
                        opts?.onStarted?.();
                },
                onData: (data) => {
                    if (!isStopped)
                        opts?.onData?.(data);
                },
                onError: (err) => {
                    if (!isStopped)
                        opts?.onError?.(err);
                },
            });
            return onDestroy(() => {
                isStopped = true;
                subscription.unsubscribe();
            });
        };
    },
    [ProcedureNames.queries]: ({ path, queriesProxy }) => {
        if (path.length !== 0)
            return;
        return (input) => {
            const proxy = queriesProxy();
            return createQueries(input(proxy));
        };
    },
    [ProcedureNames.serverQueries]: ({ path, queriesProxy, queryClient }) => {
        if (path.length !== 0)
            return;
        const proxy = queriesProxy();
        const defaultOptions = queryClient.getDefaultOptions();
        return async (input) => {
            const queries = await Promise.all(input(proxy).map(async (query) => {
                const cache = queryClient
                    .getQueryCache()
                    .find({ queryKey: query.queryKey });
                const cacheNotFound = !cache?.state?.data;
                if (query.ssr !== false && cacheNotFound) {
                    await queryClient.prefetchQuery(query);
                }
                return {
                    ...query,
                    ...(cacheNotFound
                        ? {
                            refetchOnMount: query.refetchOnMount ??
                                defaultOptions.queries?.refetchOnMount ??
                                false,
                        }
                        : {}),
                };
            }));
            return (newInput) => {
                let newQueries = queries;
                if (newInput)
                    newQueries = newInput(proxy, queries);
                return createQueries({ queries: newQueries });
            };
        };
    },
    [ProcedureNames.utils]: ({ path, utilsProxy }) => {
        if (path.length !== 0)
            return;
        return utilsProxy;
    },
    [ProcedureNames.context]: ({ path, utilsProxy }) => {
        if (path.length !== 0)
            return;
        return utilsProxy;
    },
};
function getArrayQueryKey(queryKey, input, type) {
    const arrayPath = (typeof queryKey === "string"
        ? queryKey === ""
            ? []
            : queryKey.split(".")
        : queryKey);
    if (!input && (!type || type === "any"))
        // for `utils.invalidate()` to match all queries (including vanilla react-query)
        // we don't want nested array if path is empty, i.e. `[]` instead of `[[]]`
        return arrayPath.length ? [arrayPath] : [];
    return [
        arrayPath,
        {
            ...(typeof input !== "undefined" && { input: input }),
            ...(type && type !== "any" && { type: type }),
        },
    ];
}
export function svelteQueryWrapper({ client, queryClient, abortOnUnmount, }) {
    const qc = queryClient ?? useQueryClient();
    return new DeepProxy({}, {
        get(_, key) {
            if (Object.hasOwn(procedures, key)) {
                const target = [...this.path].reduce((client, value) => client[value], client);
                return procedures[key]({
                    path: this.path,
                    target,
                    queryClient: qc,
                    queriesProxy: () => createQueriesProxy(client),
                    utilsProxy: () => createUtilsProxy(client, qc),
                    abortOnUnmount,
                });
            }
            return this.nest(() => { });
        },
    });
}
